import asyncio
import dataclasses
import itertools
import json
import logging
import os
import re
import sys
import time
import webbrowser
from collections.abc import Mapping, Callable
from pathlib import Path
from typing import NoReturn, TextIO

from databricks.sdk.service.sql import CreateWarehouseRequestWarehouseType
from databricks.sdk import WorkspaceClient

from databricks.labs.blueprint.cli import App
from databricks.labs.blueprint.entrypoint import get_logger
from databricks.labs.blueprint.installation import RootJsonValue, JsonObject, JsonValue
from databricks.labs.blueprint.tui import Prompts


from databricks.labs.lakebridge.assessments.configure_assessment import create_assessment_configurator
from databricks.labs.lakebridge.assessments import PROFILER_SOURCE_SYSTEM, PRODUCT_NAME
from databricks.labs.lakebridge.assessments.profiler import Profiler

from databricks.labs.lakebridge.config import TranspileConfig, LSPConfigOptionV1
from databricks.labs.lakebridge.contexts.application import ApplicationContext
from databricks.labs.lakebridge.connections.credential_manager import cred_file, create_credential_manager
from databricks.labs.lakebridge.connections.database_manager import DatabaseManager
from databricks.labs.lakebridge.connections.env_getter import EnvGetter
from databricks.labs.lakebridge.connections.synapse_connection_helpers import validate_synapse_pools
from databricks.labs.lakebridge.helpers.recon_config_utils import ReconConfigPrompts
from databricks.labs.lakebridge.helpers.telemetry_utils import make_alphanum_or_semver
from databricks.labs.lakebridge.reconcile.runner import ReconcileRunner
from databricks.labs.lakebridge.lineage import lineage_generator
from databricks.labs.lakebridge.reconcile.recon_config import RECONCILE_OPERATION_NAME, AGG_RECONCILE_OPERATION_NAME
from databricks.labs.lakebridge.transpiler.describe import TranspilersDescription
from databricks.labs.lakebridge.transpiler.execute import transpile as do_transpile
from databricks.labs.lakebridge.transpiler.lsp.lsp_engine import LSPEngine
from databricks.labs.lakebridge.transpiler.repository import TranspilerRepository
from databricks.labs.lakebridge.transpiler.sqlglot.sqlglot_engine import SqlglotEngine
from databricks.labs.lakebridge.transpiler.switch_runner import SwitchRunner
from databricks.labs.lakebridge.transpiler.transpile_engine import TranspileEngine

from databricks.labs.lakebridge.transpiler.transpile_status import ErrorSeverity
from databricks.labs.switch.lsp import get_switch_dialects


# Subclass to allow controlled access to protected methods.
class Lakebridge(App):

    def create_workspace_client(self) -> WorkspaceClient:
        """Create a workspace client, with the appropriate product and version information.

        This is intended only for use by the install/uninstall hooks.
        """
        self._patch_databricks_host()
        return self._workspace_client()

    def _log_level(self, raw: str) -> int:
        """Convert the log-level provided by the Databricks CLI into a logging level supported by Python."""
        log_level = super()._log_level(raw)
        # Due to an issue in the handoff of the intended logging level from the Databricks CLI to our
        # application, we can't currently distinguish between --log-level=WARN and nothing at all, where we
        # prefer (and the application logging expects) INFO.
        #
        # Rather than default to only have WARNING logs show, it's preferable to default to INFO and have
        # --log-level=WARN not work for now.
        #
        # See: https://github.com/databrickslabs/lakebridge/issues/2167
        # TODO: Remove this once #2167 has been resolved.
        if log_level == logging.WARNING:
            log_level = logging.INFO
        return log_level


lakebridge = Lakebridge(__file__)
logger = get_logger(__file__)


def raise_validation_exception(msg: str) -> NoReturn:
    raise ValueError(msg)


def _create_warehouse(ws: WorkspaceClient) -> str:

    dbsql = ws.warehouses.create_and_wait(
        name=f"lakebridge-warehouse-{time.time_ns()}",
        warehouse_type=CreateWarehouseRequestWarehouseType.PRO,
        cluster_size="Small",  # Adjust size as needed
        auto_stop_mins=30,  # Auto-stop after 30 minutes of inactivity
        enable_serverless_compute=True,
        max_num_clusters=1,
    )

    if dbsql.id is None:
        raise RuntimeError(f"Failed to create warehouse {dbsql.name}")

    logger.info(f"Created warehouse with id: {dbsql.id}")
    return dbsql.id


def _remove_warehouse(ws: WorkspaceClient, warehouse_id: str):
    ws.warehouses.delete(warehouse_id)
    logger.info(f"Removed warehouse post installation with id: {warehouse_id}")


@lakebridge.command
def transpile(  # pylint: disable=too-many-arguments
    *,
    w: WorkspaceClient,
    transpiler_config_path: str | None = None,
    source_dialect: str | None = None,
    overrides_file: str | None = None,
    target_technology: str | None = None,
    input_source: str | None = None,
    output_folder: str | None = None,
    error_file_path: str | None = None,
    skip_validation: str | None = None,
    catalog_name: str | None = None,
    schema_name: str | None = None,
    ctx: ApplicationContext | None = None,
    transpiler_repository: TranspilerRepository = TranspilerRepository.user_home(),
):
    """Transpiles source dialect to databricks dialect"""
    if ctx is None:
        ctx = ApplicationContext(w)
    del w
    logger.debug(f"Preconfigured transpiler config: {ctx.transpile_config!r}")
    ctx.add_user_agent_extra("cmd", "execute-transpile")
    checker = _TranspileConfigChecker(ctx.transpile_config, ctx.prompts, transpiler_repository)
    checker.use_transpiler_config_path(transpiler_config_path)
    checker.use_source_dialect(source_dialect)
    checker.use_overrides_file(overrides_file)
    checker.use_target_technology(target_technology)
    checker.use_input_source(input_source)
    checker.use_output_folder(output_folder)
    checker.use_error_file_path(error_file_path)
    checker.use_skip_validation(skip_validation)
    checker.use_catalog_name(catalog_name)
    checker.use_schema_name(schema_name)
    config, engine = checker.check()
    logger.debug(f"Final configuration for transpilation: {config!r}")
    _add_user_agent_extras_transpile(ctx, config, engine, transpiler_repository)
    result = asyncio.run(_transpile(ctx, config, engine))
    # DO NOT Modify this print statement, it is used by the CLI to display results in GO Table Template
    print(json.dumps(result))


def _add_user_agent_extras_transpile(
    ctx: ApplicationContext,
    config: TranspileConfig,
    engine: TranspileEngine,
    transpiler_repository: TranspilerRepository,
) -> None:
    assert config.source_dialect is not None, "Source dialect has been validated by this point."
    ctx.add_user_agent_extra("transpiler_source_tech", make_alphanum_or_semver(config.source_dialect))

    plugin_name = engine.transpiler_name
    plugin_name = re.sub(r"\s+", "_", plugin_name)
    ctx.add_user_agent_extra("transpiler_plugin_name", plugin_name)

    config_path = config.transpiler_config_path_parsed
    assert config_path is not None, "Transpiler config path has been validated by this point."
    transpiler_version = transpiler_repository.get_installed_version_given_config_path(config_path)
    if transpiler_version:
        ctx.add_user_agent_extra("transpiler_plugin_version", transpiler_version)
    else:
        logger.warning("Cannot determine transpiler plugin version.")

    # Send telemetry
    user = ctx.current_user
    logger.debug(f"User: {user}")


class _TranspileConfigChecker:
    """Helper class for the 'transpile' command to check and consolidate the configuration."""

    #
    # Configuration parameters can come from 3 sources:
    #  - Command-line arguments (e.g., --input-source, --output-folder, etc.)
    #  - The configuration file, stored in the user's workspace home directory.
    #  - User prompts.
    #
    # The conventions are:
    #  - Command-line arguments take precedence over the configuration file.
    #  - Prompting is a last resort, only used when a required configuration value has not been provided and does not
    #    have a default value.
    #  - An invalid value results in a halt, with the error message indicating the source of the invalid value. We do
    #    NOT attempt to recover from invalid values by looking for another source:
    #     - Prompting unexpectedly will break scripting and automation.
    #     - Using an alternate value will need to confusion because the behaviour will not be what the user expects.
    #
    # This ensures that we distinguish between:
    #  - Invalid command-line arguments:
    #    Resolution: fix the command-line argument value.
    #  - Invalid prompt responses:
    #    Resolution: provide a valid response to the prompt.
    #  - Invalid configuration file values:
    #    Resolution: fix the configuration file value, or provide the command-line argument to override it.
    #
    # Implementation details:
    #  - For command-line arguments and prompted values, we:
    #     - Log the raw values (prior to validation) at DEBUG level, using the repr() rendering.
    #     - Validate the values immediately, with the error message on failure mentioning the source of the value.
    #     - Only update the configuration if the validation passes.
    #  - Prompting only occurs when a value is required, but not provided via the command-line argument or the
    #    configuration file.
    #  - In addition to the above, a final validation of everything is required: this ensures that values from the
    #    configuration file are validated, and if we have a failure we know that's the source because other sources
    #    were already checked.
    #  - The interplay between the source dialect and the transpiler config path is handled with care:
    #      - The source dialect, needs to be consistent with the engine that transpiler config path, refers to.
    #      - The source dialect can be used to infer the transpiler config path.
    #
    # TODO: Refactor this class to eliminate a lof of the boilerplate and handle this more elegantly.

    _config: TranspileConfig
    """The workspace configuration for transpiling, updated from command-line arguments."""
    _prompts: Prompts
    """Prompting system, for requesting configuration that hasn't been provided."""
    _source_dialect_override: str | None = None
    """The source dialect provided on the command-line, if any."""
    _transpiler_repository: TranspilerRepository
    """The repository where available transpilers are installed."""

    def __init__(
        self,
        config: TranspileConfig | None,
        prompts: Prompts,
        transpiler_repository: TranspilerRepository,
    ) -> None:
        if config is None:
            logger.debug("No workspace transpile configuration, starting from defaults.")
            config = TranspileConfig()
        self._config = config
        self._prompts = prompts
        self._transpiler_repository = transpiler_repository
        self._source_dialect_override = None

    @staticmethod
    def _validate_transpiler_config_path(transpiler_config_path: str, msg: str) -> None:
        """Validate the transpiler config path: it must be a valid path that exists."""
        # Note: the content is not validated here, but during loading of the engine.
        if not Path(transpiler_config_path).exists():
            raise_validation_exception(msg)

    def use_transpiler_config_path(self, transpiler_config_path: str | None) -> None:
        if transpiler_config_path is not None:
            logger.debug(f"Setting transpiler_config_path to: {transpiler_config_path!r}")
            self._validate_transpiler_config_path(
                transpiler_config_path,
                f"Invalid path for '--transpiler-config-path', does not exist: {transpiler_config_path}",
            )
            self._config = dataclasses.replace(self._config, transpiler_config_path=transpiler_config_path)

    def use_source_dialect(self, source_dialect: str | None) -> None:
        if source_dialect is not None:
            # Defer validation: depends on the transpiler config path, we'll deal with this later.
            logger.debug(f"Pending source_dialect override: {source_dialect!r}")
            self._source_dialect_override = source_dialect

    @staticmethod
    def _validate_overrides_file(overrides_file: str, msg: str) -> None:
        """Validate the overrides file: it must be a valid path that exists."""
        # Note: in addition to this check, later we verify the transpiler supports it.
        if not Path(overrides_file).exists():
            raise_validation_exception(msg)

    def use_overrides_file(self, overrides_file: str | None) -> None:
        if overrides_file is not None:
            logger.debug(f"Setting overrides_file to: {overrides_file!r}")
            msg = f"Invalid path for '--overrides-file', does not exist: {overrides_file}"
            self._validate_overrides_file(overrides_file, msg)
            try:
                self._set_config_transpiler_option("overrides-file", overrides_file)
            except ValueError:
                # TODO: Update the `config.yml` format to disallow incompatible `transpiler_options`.
                msg = "Cannot use --overrides-file; workspace config.yml has incompatible transpiler_options."
                raise_validation_exception(msg)

    def use_target_technology(self, target_technology: str | None) -> None:
        if target_technology is not None:
            logger.debug(f"Setting target_technology to: {target_technology!r}")
            # Cannot validate this here: depends on the transpiler engine, and will be checked later.
            try:
                self._set_config_transpiler_option("target-tech", target_technology)
            except ValueError:
                # TODO: Update the `config.yml` format to disallow incompatible `transpiler_options`.
                msg = "Cannot use --target-technology; workspace config.yml has incompatible transpiler_options."
                raise_validation_exception(msg)

    @staticmethod
    def _validate_input_source(input_source: str, msg: str) -> None:
        """Validate the input source: it must be a path that exists."""
        if not Path(input_source).exists():
            raise_validation_exception(msg)

    def use_input_source(self, input_source: str | None) -> None:
        if input_source is not None:
            logger.debug(f"Setting input_source to: {input_source!r}")
            self._validate_input_source(
                input_source, f"Invalid path for '--input-source', does not exist: {input_source}"
            )
            self._config = dataclasses.replace(self._config, input_source=input_source)

    def _prompt_input_source(self) -> None:
        prompted_input_source = self._prompts.question("Enter input SQL path (directory/file)").strip()
        logger.debug(f"Setting input_source to: {prompted_input_source!r}")
        self._validate_input_source(
            prompted_input_source, f"Invalid input source, path does not exist: {prompted_input_source}"
        )
        self._config = dataclasses.replace(self._config, input_source=prompted_input_source)

    def _check_input_source(self) -> None:
        config_input_source = self._config.input_source
        if config_input_source is None:
            self._prompt_input_source()
        else:
            self._validate_input_source(
                config_input_source, f"Invalid input source path configured, does not exist: {config_input_source}"
            )

    @staticmethod
    def _validate_output_folder(output_folder: str, msg: str) -> None:
        """Validate the output folder: it doesn't have to exist, but its parent must."""
        if not Path(output_folder).parent.exists():
            raise_validation_exception(msg)

    def use_output_folder(self, output_folder: str | None) -> None:
        if output_folder is not None:
            logger.debug(f"Setting output_folder to: {output_folder!r}")
            self._validate_output_folder(
                output_folder, f"Invalid path for '--output-folder', parent does not exist for: {output_folder}"
            )
            self._config = dataclasses.replace(self._config, output_folder=output_folder)

    def _prompt_output_folder(self) -> None:
        prompted_output_folder = self._prompts.question("Enter output folder path (directory)").strip()
        logger.debug(f"Setting output_folder to: {prompted_output_folder!r}")
        self._validate_output_folder(
            prompted_output_folder, f"Invalid output folder path, parent does not exist for: {prompted_output_folder}"
        )
        self._config = dataclasses.replace(self._config, output_folder=prompted_output_folder)

    def _check_output_folder(self) -> None:
        config_output_folder = self._config.output_folder
        if config_output_folder is None:
            self._prompt_output_folder()
        else:
            self._validate_output_folder(
                config_output_folder,
                f"Invalid output folder configured, parent does not exist for: {config_output_folder}",
            )

    @staticmethod
    def _validate_error_file_path(error_file_path: str | None, msg: str) -> None:
        """Value the error file path: it doesn't have to exist, but its parent must."""
        if error_file_path is not None and not Path(error_file_path).parent.exists():
            raise_validation_exception(msg)

    def use_error_file_path(self, error_file_path: str | None) -> None:
        if error_file_path is not None:
            logger.debug(f"Setting error_file_path to: {error_file_path!r}")
            self._validate_error_file_path(
                error_file_path, f"Invalid path for '--error-file-path', parent does not exist: {error_file_path}"
            )
            self._config = dataclasses.replace(self._config, error_file_path=error_file_path)

    def _check_error_file_path(self) -> None:
        config_error_file_path = self._config.error_file_path
        self._validate_error_file_path(
            config_error_file_path,
            f"Invalid error file path configured, parent does not exist for: {config_error_file_path}",
        )

    def use_skip_validation(self, skip_validation: str | None) -> None:
        if skip_validation is not None:
            skip_validation_lower = skip_validation.lower()
            if skip_validation_lower not in {"true", "false"}:
                msg = f"Invalid value for '--skip-validation': {skip_validation!r} must be 'true' or 'false'."
                raise_validation_exception(msg)
            new_skip_validation = skip_validation_lower == "true"
            logger.debug(f"Setting skip_validation to: {new_skip_validation!r}")
            self._config = dataclasses.replace(self._config, skip_validation=new_skip_validation)

    def use_catalog_name(self, catalog_name: str | None) -> None:
        if catalog_name:
            logger.debug(f"Setting catalog_name to: {catalog_name!r}")
            self._config = dataclasses.replace(self._config, catalog_name=catalog_name)

    def use_schema_name(self, schema_name: str | None) -> None:
        if schema_name:
            logger.debug(f"Setting schema_name to: {schema_name!r}")
            self._config = dataclasses.replace(self._config, schema_name=schema_name)

    def _set_config_transpiler_option(self, flag: str, value: str) -> None:
        transpiler_options: JsonObject
        match self._config.transpiler_options:
            case None:
                transpiler_options = {flag: value}
            case Mapping() as found_options:
                transpiler_options = {**found_options, flag: value}
            case found_options:
                # TODO: Update `config.yml' to constrain `transpiler_options` to be a dict[str, str].
                msg = f"Incompatible transpiler options configured, must be a mapping: {found_options!r}"
                raise ValueError(msg)
        self._config = dataclasses.replace(self._config, transpiler_options=transpiler_options)

    def _configure_transpiler_config_path(self, source_dialect: str) -> TranspileEngine | None:
        """Configure the transpiler config path based on the requested source dialect."""
        # Names of compatible transpiler engines for the given dialect.
        compatible_transpilers = self._transpiler_repository.transpilers_with_dialect(source_dialect)
        match len(compatible_transpilers):
            case 0:
                # Nothing found for the specified dialect, fail.
                return None
            case 1:
                # Only one transpiler available for the specified dialect, use it.
                transpiler_name = next(iter(compatible_transpilers))
                logger.debug(f"Using only transpiler available for dialect {source_dialect!r}: {transpiler_name!r}")
            case _:
                # Multiple transpilers available for the specified dialect, prompt for which to use.
                logger.debug(
                    f"Multiple transpilers available for dialect {source_dialect!r}: {compatible_transpilers!r}"
                )
                transpiler_name = self._prompts.choice("Select the transpiler:", list(compatible_transpilers))
        transpiler_config_path = self._transpiler_repository.transpiler_config_path(transpiler_name)
        logger.info(f"Lakebridge will use the {transpiler_name} transpiler.")
        self._config = dataclasses.replace(self._config, transpiler_config_path=str(transpiler_config_path))
        return LSPEngine.from_config_path(transpiler_config_path)

    def _configure_source_dialect(
        self, source_dialect: str, engine: TranspileEngine | None, msg_prefix: str
    ) -> TranspileEngine:
        """Configure the source dialect, if possible, and return the transpiler engine."""
        if engine is None:
            engine = self._configure_transpiler_config_path(source_dialect)
            if engine is None:
                supported_dialects = ", ".join(self._transpiler_repository.all_dialects())
                msg = f"{msg_prefix}: {source_dialect!r} (supported dialects: {supported_dialects})"
                raise_validation_exception(msg)
            else:
                self._config = dataclasses.replace(self._config, source_dialect=source_dialect)
        else:
            # Check the source dialect against the engine.
            if source_dialect not in engine.supported_dialects:
                supported_dialects_description = ", ".join(engine.supported_dialects)
                msg = f"Invalid value for '--source-dialect': {source_dialect!r} must be one of: {supported_dialects_description}"
                raise_validation_exception(msg)
            self._config = dataclasses.replace(self._config, source_dialect=source_dialect)
        return engine

    def _prompt_source_dialect(self) -> TranspileEngine:
        # This is similar to the post-install prompting for the source dialect.
        supported_dialects = self._transpiler_repository.all_dialects()
        match len(supported_dialects):
            case 0:
                msg = "No transpilers are available, install using 'install-transpile' or use --transpiler-conf-path'."
                raise_validation_exception(msg)
            case 1:
                # Only one dialect available, use it.
                source_dialect = next(iter(supported_dialects))
                logger.debug(f"Using only source dialect available: {source_dialect!r}")
            case _:
                # Multiple dialects available, prompt for which to use.
                logger.debug(f"Multiple source dialects available, choice required: {supported_dialects!r}")
                source_dialect = self._prompts.choice("Select the source dialect:", list(supported_dialects))
        engine = self._configure_transpiler_config_path(source_dialect)
        assert engine is not None, "No transpiler engine available for a supported dialect; configuration is invalid."
        self._config = dataclasses.replace(self._config, source_dialect=source_dialect)
        return engine

    def _check_lsp_engine(self) -> TranspileEngine:
        #
        # This is somewhat complicated:
        #  - If there is no transpiler config path, we need to try to infer it from the source dialect.
        #  - If there is no source dialect, we need to prompt for it: but that depends on the transpiler config path.
        #
        # With this in mind, the steps here are:
        # 1. If the transpiler config path is set, check it exists and load the engine.
        # 2. If the source dialect is set,
        #      - If the transpiler config path is set: validate the source dialect against the engine.
        #      - If the transpiler config path is not set: search for a transpiler that satisfies the dialect:
        #          * If one is found, we're good to go.
        #          * If more than one is found, prompt for the transpiler config path.
        #          * If none are found, fail: no transpilers available for the specified dialect.
        #    At this point we have either halted, or we have a valid transpiler path and source dialect.
        # 3. If the source dialect is not set, we need to:
        #      a) Load the set of available dialects: just for the engine if transpiler config path is set, or for all
        #         available transpilers if not.
        #      b) Depending on the available dialects:
        #          - If there is only one dialect available, set it as the source dialect.
        #          - If there are multiple dialects available, prompt for which to use.
        #          - If there are no dialects available, fail: no transpilers available.
        #    At this point we have either halted, or we have a valid transpiler path and source dialect.
        #
        # TODO: Deal with the transpiler options, and filtering them for the engine.
        #

        # Step 1: Check the transpiler config path.
        engine: TranspileEngine | None
        transpiler_config_path = self._config.transpiler_config_path
        if transpiler_config_path is not None:
            self._validate_transpiler_config_path(
                transpiler_config_path,
                f"Error: Invalid value for '--transpiler-config-path': '{str(transpiler_config_path)}', file does not exist.",
            )
            path = Path(transpiler_config_path)
            engine = LSPEngine.from_config_path(path)
        else:
            engine = None
        del transpiler_config_path

        # Step 2: Check the source dialect, assuming it has been specified, and infer the transpiler config path if necessary.
        source_dialect = self._source_dialect_override
        if source_dialect is not None:
            logger.debug(f"Setting source_dialect override: {source_dialect!r}")
            engine = self._configure_source_dialect(source_dialect, engine, "Invalid value for '--source-dialect'")
        else:
            source_dialect = self._config.source_dialect
            if source_dialect is not None:
                logger.debug(f"Using configured source_dialect: {source_dialect!r}")
                engine = self._configure_source_dialect(source_dialect, engine, "Invalid configured source dialect")
            else:
                # Step 3: Source dialect is not set, we need to prompt for it.
                logger.debug("No source_dialect available, prompting.")
                engine = self._prompt_source_dialect()
        return engine

    def _check_transpiler_options(self, engine: TranspileEngine) -> None:
        if not isinstance(engine, LSPEngine):
            return
        assert self._config.source_dialect is not None, "Source dialect must be set before checking transpiler options."
        options_for_dialect = engine.options_for_dialect(self._config.source_dialect)
        transpiler_options = self._config.transpiler_options
        if transpiler_options is None:
            transpiler_options = {}
        elif not isinstance(transpiler_options, Mapping):
            logger.warning(f"Ignoring transpiler_options in config.yml, must be a mapping: {transpiler_options!r}")
            transpiler_options = {}
        # Only checks if the option is present, does not validate the value.
        # TODO: Validate the value for CHOICE/CONFIRM options.
        # TODO: Handle FORCE options: these are fixed by the transpiler, and cannot be overridden.
        checked_options = {
            option.flag: (
                transpiler_options[option.flag]
                if option.flag in transpiler_options
                else self._handle_missing_transpiler_option(option)
            )
            for option in options_for_dialect
        }
        self._config = dataclasses.replace(self._config, transpiler_options=checked_options)

    def _handle_missing_transpiler_option(self, option: LSPConfigOptionV1) -> JsonValue:
        # Semantics during configuration:
        #  - Entries are present in the config file for all options the LSP server needs for a dialect.
        #  - If a value is `None`, it means the user wants the value to be left unset.
        #  - There is no 'provide it later' option: either it's set, or it's unset.
        # As a corner case, if there is no entry present it means the user wasn't prompted. Here we have
        # some complexity. We have two ways of obtaining a value:
        #  - The user could provide it on the command-line, using --target-technology or --overrides-file.
        #    Problem: via command-line options there's no way to indicate 'no value'.
        #  - We could prompt for it, assuming the user is running interactively.
        # In terms of what is required by the option:
        #  - If the option has a default of <none>, it means that no value is required.
        #  - Everything else requires a value.
        #
        # This leads to the following business rules:
        #  - If the option has a default of <none> that means that no value is required, no further action is required.
        #  - Otherwise, a value is required: prompt for it.
        #
        # TODO: When adding non-interactive support, the otherwise branch need to be modified:
        #     1. If it can be provided by the command-line, fail and ask the user to provide it.
        #     2. If it cannot be provided by the command-line, prompt for it if we are running interactively.
        #     3. If we cannot prompt because we are not running interactively, use the default if there is one.
        #     4. Fail: the only way to provide a value is via the config.yml, which can be set via 'install-transpile'.

        if option.is_optional():
            return None
        return option.prompt_for_value(self._prompts)

    def check(self) -> tuple[TranspileConfig, TranspileEngine]:
        """Checks that all configuration parameters are present and valid."""
        logger.debug(f"Checking config: {self._config!r}")

        self._check_input_source()
        self._check_output_folder()
        self._check_error_file_path()
        # No validation here required for:
        #   - skip_validation: it is a boolean flag, mandatory, and has a default: so no further validation is needed.
        #   - catalog_name and schema_name: they are mandatory, but have a default.
        # TODO: if validation is enabled, we should check that the catalog and schema names are valid.

        # This covers: transpiler_config_path, source_dialect
        engine = self._check_lsp_engine()

        # Last thing: the configuration may have transpiler-specific options, check them.
        self._check_transpiler_options(engine)

        config = self._config
        logger.debug(f"Validated config: {config!r}")
        return config, engine


async def _transpile(ctx: ApplicationContext, config: TranspileConfig, engine: TranspileEngine) -> RootJsonValue:
    """Transpiles source dialect to databricks dialect"""
    _override_workspace_client_config(ctx, config.sdk_config)
    status, errors = await do_transpile(ctx.workspace_client, engine, config)

    logger.debug(f"Transpilation completed with status: {status}")

    for path, errors_by_path in itertools.groupby(errors, key=lambda x: x.path):
        errs = list(errors_by_path)
        errors_by_severity = {
            severity.name: len(list(errors)) for severity, errors in itertools.groupby(errs, key=lambda x: x.severity)
        }
        reports = []
        reported_severities = [ErrorSeverity.ERROR, ErrorSeverity.WARNING]
        for severity in reported_severities:
            if severity.name in errors_by_severity:
                word = str.lower(severity.name) + "s" if errors_by_severity[severity.name] > 1 else ""
                reports.append(f"{errors_by_severity[severity.name]} {word}")

        msg = ", ".join(reports) + " found"

        if ErrorSeverity.ERROR.name in errors_by_severity:
            logger.error(f"{path}: {msg}")
        elif ErrorSeverity.WARNING.name in errors_by_severity:
            logger.warning(f"{path}: {msg}")

    # Table Template in labs.yml requires the status to be list of dicts Do not change this
    return [status]


def _override_workspace_client_config(ctx: ApplicationContext, overrides: dict[str, str] | None) -> None:
    """
    Override the Workspace client's SDK config with the user provided SDK config.
    Users can provide the cluster_id and warehouse_id during the installation.
    This will update the default config object in-place.
    """
    if not overrides:
        return

    warehouse_id = overrides.get("warehouse_id")
    if warehouse_id:
        ctx.connect_config.warehouse_id = warehouse_id

    cluster_id = overrides.get("cluster_id")
    if cluster_id:
        ctx.connect_config.cluster_id = cluster_id


@lakebridge.command
def reconcile(
    *, w: WorkspaceClient, ctx_factory: Callable[[WorkspaceClient], ApplicationContext] = ApplicationContext
) -> None:
    """[EXPERIMENTAL] Reconciles source to Databricks datasets"""
    ctx = ctx_factory(w)
    ctx.add_user_agent_extra("cmd", "execute-reconcile")
    user = ctx.current_user
    logger.debug(f"User: {user}")
    recon_runner = ReconcileRunner(
        ctx.workspace_client,
        ctx.install_state,
    )

    _, job_run_url = recon_runner.run(operation_name=RECONCILE_OPERATION_NAME)
    if ctx.prompts.confirm(f"Would you like to open the job run URL `{job_run_url}` in the browser?"):
        webbrowser.open(job_run_url)


@lakebridge.command
def aggregates_reconcile(
    *, w: WorkspaceClient, ctx_factory: Callable[[WorkspaceClient], ApplicationContext] = ApplicationContext
) -> None:
    """[EXPERIMENTAL] Reconciles Aggregated source to Databricks datasets"""
    ctx = ctx_factory(w)
    ctx.add_user_agent_extra("cmd", "execute-aggregates-reconcile")
    user = ctx.current_user
    logger.debug(f"User: {user}")
    recon_runner = ReconcileRunner(
        ctx.workspace_client,
        ctx.install_state,
    )

    _, job_run_url = recon_runner.run(operation_name=AGG_RECONCILE_OPERATION_NAME)
    if ctx.prompts.confirm(f"Would you like to open the job run URL `{job_run_url}` in the browser?"):
        webbrowser.open(job_run_url)


@lakebridge.command
def generate_lineage(
    *,
    w: WorkspaceClient,
    source_dialect: str | None = None,
    input_source: str,
    output_folder: str,
) -> None:
    """[Experimental] Generates a lineage of source SQL files or folder"""
    ctx = ApplicationContext(w)
    logger.debug(f"User: {ctx.current_user}")
    if not os.path.exists(input_source):
        raise_validation_exception(f"Invalid path for '--input-source': Path '{input_source}' does not exist.")
    if not os.path.exists(output_folder):
        raise_validation_exception(f"Invalid path for '--output-folder': Path '{output_folder}' does not exist.")
    if source_dialect is None:
        raise_validation_exception("Value for '--source-dialect' must be provided.")
    engine = SqlglotEngine()
    supported_dialects = engine.supported_dialects
    if source_dialect not in supported_dialects:
        supported_dialects_description = ", ".join(supported_dialects)
        msg = f"Unsupported source dialect provided for '--source-dialect': '{source_dialect}' (supported: {supported_dialects_description})"
        raise_validation_exception(msg)

    lineage_generator(engine, source_dialect, input_source, output_folder)


@lakebridge.command
def configure_secrets(*, w: WorkspaceClient) -> None:
    """Setup reconciliation connection profile details as Secrets on Databricks Workspace"""
    recon_conf = ReconConfigPrompts(w)

    # Prompt for source
    source = recon_conf.prompt_source()

    logger.info(f"Setting up Scope, Secrets for `{source}` reconciliation")
    recon_conf.prompt_and_save_connection_details()


@lakebridge.command
def configure_database_profiler(w: WorkspaceClient) -> None:
    """[Experimental] Installs and runs the Lakebridge Assessment package for database profiling"""
    ctx = ApplicationContext(w)
    ctx.add_user_agent_extra("cmd", "configure-profiler")
    prompts = ctx.prompts
    source_tech = prompts.choice("Select the source technology", PROFILER_SOURCE_SYSTEM).lower()
    ctx.add_user_agent_extra("profiler_source_tech", make_alphanum_or_semver(source_tech))
    user = ctx.current_user
    logger.debug(f"User: {user}")

    # Create appropriate assessment configurator
    assessment = create_assessment_configurator(source_system=source_tech, product_name="lakebridge", prompts=prompts)
    assessment.run()


@lakebridge.command
def install_transpile(
    *,
    w: WorkspaceClient,
    artifact: str | None = None,
    interactive: str | None = None,
    include_llm_transpiler: bool = False,
    transpiler_repository: TranspilerRepository = TranspilerRepository.user_home(),
) -> None:
    """Install or upgrade the Lakebridge transpilers."""
    # Avoid circular imports.
    from databricks.labs.lakebridge.install import installer  # pylint: disable=cyclic-import, import-outside-toplevel

    is_interactive = interactive_mode(interactive)
    ctx = ApplicationContext(w)
    ctx.add_user_agent_extra("cmd", "install-transpile")
    if artifact:
        ctx.add_user_agent_extra("artifact-overload", Path(artifact).name)
    # Internal: use LAKEBRIDGE_CLUSTER_TYPE=CLASSIC env var to use classic job cluster
    switch_use_serverless = os.environ.get("LAKEBRIDGE_CLUSTER_TYPE", "").upper() != "CLASSIC"
    if include_llm_transpiler:
        ctx.add_user_agent_extra("include-llm-transpiler", "true")
        # Decision was made not to prompt when include_llm_transpiler is set, and we expect users to use llm-transpile
        # and pass all the arguments.
        logger.info("Including LLM transpiler as part of install, interactive mode disabled: will skip questionnaire.")
        is_interactive = False

    user = w.current_user
    logger.debug(f"User: {user}")
    transpile_installer = installer(
        w,
        transpiler_repository,
        is_interactive=is_interactive,
        include_llm=include_llm_transpiler,
        switch_use_serverless=switch_use_serverless,
    )
    transpile_installer.run(module="transpile", artifact=artifact)


def interactive_mode(interactive: str | None, *, default: str = "auto", input_stream: TextIO = sys.stdin) -> bool:
    """Convert the raw '--interactive' argument into a boolean."""
    if interactive is None:
        interactive = default
    match interactive.lower():
        case "true":
            return True
        case "false":
            return False
        # Convention is that if the input_stream is a TTY, user interaction is allowed.
        case "auto":
            return input_stream.isatty()

    msg = f"Invalid value for '--interactive': {interactive!r} must be 'true', 'false' or 'auto'."
    raise_validation_exception(msg)


@lakebridge.command
def describe_transpile(
    *,
    w: WorkspaceClient,
    transpiler_repository: TranspilerRepository = TranspilerRepository.user_home(),
) -> None:
    """Describe the installed Lakebridge transpilers and available options."""
    ctx = ApplicationContext(w)
    ctx.add_user_agent_extra("cmd", "describe-transpile")
    user = w.current_user.me()
    logger.debug(f"User: {user}")
    transpilers_description = TranspilersDescription(transpiler_repository)
    json_description = transpilers_description.as_json()
    json.dump(json_description, sys.stdout, indent=2)


@lakebridge.command(is_unauthenticated=False)
def configure_reconcile(
    *,
    w: WorkspaceClient,
    transpiler_repository: TranspilerRepository = TranspilerRepository.user_home(),
) -> None:
    """Configure the Lakebridge reconciliation module"""
    # Avoid circular imports.
    from databricks.labs.lakebridge.install import installer  # pylint: disable=cyclic-import, import-outside-toplevel

    ctx = ApplicationContext(w)
    ctx.add_user_agent_extra("cmd", "configure-reconcile")
    user = w.current_user
    logger.debug(f"User: {user}")
    if not w.config.warehouse_id:
        dbsql_id = _create_warehouse(w)
        w.config.warehouse_id = dbsql_id
    logger.debug(f"Warehouse ID used for configuring reconcile: {w.config.warehouse_id}.")
    reconcile_installer = installer(w, transpiler_repository, is_interactive=True)
    reconcile_installer.run(module="reconcile")


@lakebridge.command
def analyze(
    *,
    w: WorkspaceClient,
    source_directory: str | None = None,
    report_file: str | None = None,
    source_tech: str | None = None,
    generate_json: bool = False,
):
    """Run the Analyzer"""
    ctx = ApplicationContext(w)
    try:
        result = ctx.analyzer.run_analyzer(source_directory, report_file, source_tech, generate_json)
        ctx.add_user_agent_extra("analyzer_source_tech", result.source_system)
    finally:
        exception_cls, _, _ = sys.exc_info()
        if exception_cls is not None:
            ctx.add_user_agent_extra("analyzer_error", exception_cls.__name__)

        ctx.add_user_agent_extra("cmd", "analyze")
        logger.debug(f"User: {ctx.current_user}")


def _validate_llm_transpile_args(
    input_source: str | None,
    output_ws_folder: str | None,
    source_dialect: str | None,
    prompts: Prompts,
) -> tuple[str, str, str]:

    _switch_dialects = get_switch_dialects()

    # Validate presence after attempting to source from config
    if not input_source:
        input_source = prompts.question("Enter input SQL path")
    if not output_ws_folder:
        output_ws_folder = prompts.question("Enter output workspace folder must start with /Workspace/")
    if not source_dialect:
        source_dialect = prompts.choice("Select the source dialect", sorted(_switch_dialects))

    # Validate input_source path exists (local path)
    if not Path(input_source).exists():
        raise_validation_exception(f"Invalid path for '--input-source': Path '{input_source}' does not exist.")

    # Validate output_ws_folder is a workspace path
    if not str(output_ws_folder).startswith("/Workspace/"):
        raise_validation_exception(
            f"Invalid value for '--output-ws-folder': workspace output path must start with /Workspace/. Got: {output_ws_folder!r}"
        )

    if source_dialect not in _switch_dialects:
        raise_validation_exception(
            f"Invalid value for '--source-dialect': {source_dialect!r} must be one of: {', '.join(sorted(_switch_dialects))}"
        )

    return input_source, output_ws_folder, source_dialect


@lakebridge.command
def llm_transpile(
    *,
    w: WorkspaceClient,
    accept_terms: bool = False,
    input_source: str | None = None,
    output_ws_folder: str | None = None,
    source_dialect: str | None = None,
    catalog_name: str | None = None,
    schema_name: str | None = None,
    volume: str | None = None,
    foundation_model: str | None = None,
    ctx: ApplicationContext | None = None,
) -> None:
    """Transpile source code to Databricks using LLM Transpiler (Switch)"""
    if ctx is None:
        ctx = ApplicationContext(w)
    del w
    ctx.add_user_agent_extra("cmd", "llm-transpile")
    user = ctx.current_user
    logger.debug(f"User: {user}")

    if not accept_terms:
        logger.warning(
            """Please read and accept these terms before proceeding:
    This feature leverages a Large Language Model (LLM) to analyse and convert
    your provided content, code and data. You consent to your content being
    transmitted to, processed by, and returned from the foundation models hosted
    by Databricks or external foundation models you have configured in your
    workspace. The outputs of the LLM are generated automatically without human
    review, and may contain inaccuracies or errors. You are responsible for
    reviewing and validating all outputs before relying on them for any critical
    or production use.

    By using this feature you accept these terms, re-run with '--accept-terms=true'.
                """
        )
        raise SystemExit("LLM transpiler terms not accepted, exiting.")

    prompts = ctx.prompts
    resource_configurator = ctx.resource_configurator

    # If CLI args are missing, try to read them from config.yml
    input_source, output_ws_folder, source_dialect = _validate_llm_transpile_args(
        input_source,
        output_ws_folder,
        source_dialect,
        prompts,
    )

    if catalog_name is None:
        catalog_name = resource_configurator.prompt_for_catalog_setup(default_catalog_name="lakebridge")

    if schema_name is None:
        schema_name = resource_configurator.prompt_for_schema_setup(catalog=catalog_name, default_schema_name="switch")

    if volume is None:
        volume = resource_configurator.prompt_for_volume_setup(
            catalog=catalog_name, schema=schema_name, default_volume_name="switch_volume"
        )

    resource_configurator.has_necessary_access(catalog_name, schema_name, volume)

    if foundation_model is None:
        foundation_model = resource_configurator.prompt_for_foundation_model_choice()

    job_list = ctx.install_state.jobs
    if "Switch" not in job_list:
        logger.debug(f"Missing Switch from installed state jobs: {job_list!r}")
        raise RuntimeError(
            "Switch Job not found. "
            "Please run 'databricks labs lakebridge install-transpile --include-llm-transpiler true' first."
        )
    job_id = int(job_list["Switch"])
    logger.debug(f"Switch job ID found: {job_id}")

    ctx.add_user_agent_extra("transpiler_source_dialect", source_dialect)
    job_runner = SwitchRunner(ctx.workspace_client)
    volume_input_path = job_runner.upload_to_volume(
        local_path=Path(input_source),
        catalog=catalog_name,
        schema=schema_name,
        volume=volume,
    )

    job_runner.run(
        volume_input_path=volume_input_path,
        output_ws_folder=output_ws_folder,
        source_tech=source_dialect,
        catalog=catalog_name,
        schema=schema_name,
        foundation_model=foundation_model,
        job_id=job_id,
    )


@lakebridge.command()
def execute_database_profiler(w: WorkspaceClient, source_tech: str | None = None) -> None:
    """Execute the Profiler Extraction for the given source technology"""
    ctx = ApplicationContext(w)
    ctx.add_user_agent_extra("cmd", "execute-profiler")
    prompts = ctx.prompts
    if source_tech is None:
        source_tech = prompts.choice("Select the source technology", PROFILER_SOURCE_SYSTEM)
    source_tech = source_tech.lower()

    if source_tech not in PROFILER_SOURCE_SYSTEM:
        logger.error(f"Only the following source systems are supported: {PROFILER_SOURCE_SYSTEM}")
        raise_validation_exception(f"Invalid source technology {source_tech}")

    ctx.add_user_agent_extra("profiler_source_tech", make_alphanum_or_semver(source_tech))
    user = ctx.current_user
    logger.debug(f"User: {user}")
    # check if cred_file is present which has the connection details before running the profiler
    file = cred_file(PRODUCT_NAME)
    if not file.exists():
        raise_validation_exception(
            f"Connection details not found. Please run `databricks labs lakebridge configure-database-profiler` "
            f"to set up connection details for {source_tech}."
        )
    profiler = Profiler.create(source_tech)

    # TODO: Add extractor logic to ApplicationContext instead of creating inside the Profiler class
    profiler.profile()


@lakebridge.command()
def create_profiler_dashboard(
    *,
    w: WorkspaceClient,
    extract_file: str,
    source_tech: str,
    volume_path: str,
    catalog_name: str,
    schema_name: str,
) -> None:
    """Deploys a profiler summary as a Databricks dashboard"""
    ctx = ApplicationContext(w)
    ctx.add_user_agent_extra("cmd", "create-profiler-dashboard")
    ctx.dashboard_manager.upload_duckdb_to_uc_volume(extract_file, volume_path)
    ctx.dashboard_manager.create_profiler_summary_dashboard(source_tech, catalog_name, schema_name)


def _test_database_connection(source_tech: str, raw_config: dict) -> None:
    """Test connection to the source database with appropriate error handling."""
    # Handle synapse-specific validation using dedicated helper
    if source_tech == "synapse":
        validate_synapse_pools(raw_config)
        logger.info("Connection to the source system successful")
        return

    # For other source technologies, use DatabaseManager directly
    with DatabaseManager(source_tech, raw_config) as db_manager:
        response = db_manager.check_connection()
    logger.debug(f"Connection response: {response}")
    logger.info("Connection to the source system successful")


@lakebridge.command()
def test_profiler_connection(
    *,
    w: WorkspaceClient,
    source_tech: str | None = None,
    cred_file_path: str | None = None,
) -> None:
    """[Internal] Test the connection to the source database for profiling"""
    ctx = ApplicationContext(w)
    ctx.add_user_agent_extra("cmd", "test-profiler-connection")
    prompts = ctx.prompts

    source_tech = (
        source_tech.lower()
        if source_tech
        else prompts.choice("Select the source technology", PROFILER_SOURCE_SYSTEM).lower()
    )

    if source_tech not in PROFILER_SOURCE_SYSTEM:
        logger.error(f"Only the following source systems are supported: {PROFILER_SOURCE_SYSTEM}")
        raise_validation_exception(f"Invalid source technology {source_tech}")

    ctx.add_user_agent_extra("profiler_source_tech", make_alphanum_or_semver(source_tech))
    logger.debug(f"User: {ctx.current_user}")

    # Use provided credential file path or fall back to default
    credential_file = Path(cred_file_path) if cred_file_path else cred_file(PRODUCT_NAME)

    # Check if credential file exists
    if not credential_file.exists():
        raise_validation_exception(
            f"Connection details not found. Please run `databricks labs lakebridge configure-database-profiler` "
            f"to set up connection details for {source_tech}."
        )

    logger.info(f"Testing connection for source technology: {source_tech}")

    cred_manager = create_credential_manager(PRODUCT_NAME, EnvGetter(), creds_path=credential_file)

    try:
        raw_config = cred_manager.get_credentials(source_tech)
    except KeyError as e:
        logger.error(f"Credential configuration error: {e}")
        logger.fatal(
            f"Invalid credentials for {source_tech}. Please run `databricks labs lakebridge configure-database-profiler`."
        )
        return

    try:
        _test_database_connection(source_tech, raw_config)
    except ConnectionError as e:
        logger.error(f"Failed to connect to the source system: {e}")
        error_msg = str(e).lower()
        if any(pattern in error_msg for pattern in ("im002", "odbc driver not found", "can't open lib")):
            logger.fatal("Missing ODBC driver, Please install pre-req. Exiting...")
        else:
            logger.fatal("Connection validation failed. Exiting...")
    except Exception as e:  # pylint: disable=broad-exception-caught
        # Catch all exceptions to provide user-friendly error messages for CLI
        logger.error(f"Unexpected error during connection test: {e}")
        logger.fatal("Connection test failed. Exiting...")


if __name__ == "__main__":
    lakebridge()



from __future__ import annotations

import logging

from dataclasses import dataclass
from collections.abc import Callable
from sqlglot import expressions as exp

from databricks.labs.lakebridge.reconcile.connectors.dialect_utils import DialectUtils
from databricks.labs.lakebridge.reconcile.constants import SamplingOptionMethod, SamplingSpecificationsType

logger = logging.getLogger(__name__)

_SUPPORTED_AGG_TYPES: set[str] = {
    "min",
    "max",
    "count",
    "sum",
    "avg",
    "mean",
    "mode",
    "stddev",
    "variance",
    "median",
}

RECONCILE_OPERATION_NAME = "reconcile"
AGG_RECONCILE_OPERATION_NAME = "aggregates-reconcile"

PrimitiveType = bool | int | float | str
OptionalPrimitiveType = PrimitiveType | None


class TableThresholdBoundsException(ValueError):
    """Raise the error when the bounds for table threshold are invalid"""


class InvalidModelForTableThreshold(ValueError):
    """Raise the error when the model for table threshold is invalid"""


@dataclass
class HashAlgoMapping:
    source: Callable
    target: Callable


@dataclass
class SamplingSpecifications:
    type: SamplingSpecificationsType
    value: float

    def __post_init__(self):
        if not isinstance(self.type, SamplingSpecificationsType):
            self.type = SamplingSpecificationsType(str(self.type).lower())
        # Disabled
        if self.type == SamplingSpecificationsType.FRACTION:
            raise ValueError("SamplingSpecifications: 'FRACTION' type is disabled")
        if self.type == SamplingSpecificationsType.FRACTION and (self.value is None or (not 0 < self.value < 1)):
            raise ValueError("SamplingSpecifications: Fraction value must be greater than  0 and less than 1")


@dataclass
class SamplingOptions:
    method: SamplingOptionMethod
    specifications: SamplingSpecifications
    stratified_columns: list[str] | None = None
    stratified_buckets: int | None = None

    def __post_init__(self):
        if not isinstance(self.method, SamplingOptionMethod):
            self.method = SamplingOptionMethod(str(self.method).lower())

        if self.stratified_columns:
            self.stratified_columns = [col.lower() for col in self.stratified_columns]

        if self.method == SamplingOptionMethod.STRATIFIED:
            if not self.stratified_columns or not self.stratified_buckets:
                raise ValueError(
                    "SamplingOptions : stratified_columns and stratified_buckets are required for STRATIFIED method"
                )


@dataclass
class JdbcReaderOptions:
    number_partitions: int | None = None
    partition_column: str | None = None
    lower_bound: str | None = None
    upper_bound: str | None = None
    fetch_size: int = 100

    def __post_init__(self):
        self.partition_column = self.partition_column.lower() if self.partition_column else None


@dataclass
class ColumnMapping:
    source_name: str
    target_name: str

    def __post_init__(self):
        self.source_name = self.source_name.lower()
        self.target_name = self.target_name.lower()


@dataclass
class Transformation:
    column_name: str
    source: str | None = None
    target: str | None = None

    def __post_init__(self):
        self.column_name = self.column_name.lower()


@dataclass
class ColumnThresholds:
    column_name: str
    lower_bound: str
    upper_bound: str
    type: str

    def __post_init__(self):
        self.column_name = self.column_name.lower()
        self.type = self.type.lower()

    def get_mode(self):
        return "percentage" if "%" in self.lower_bound or "%" in self.upper_bound else "absolute"

    def get_type(self):
        if any(self.type in numeric_type.value.lower() for numeric_type in exp.DataType.NUMERIC_TYPES):
            if self.get_mode() == "absolute":
                return "number_absolute"
            return "number_percentage"

        if any(self.type in numeric_type.value.lower() for numeric_type in exp.DataType.TEMPORAL_TYPES):
            return "datetime"
        return None


@dataclass
class TableThresholds:
    lower_bound: str
    upper_bound: str
    model: str

    def __post_init__(self):
        self.model = self.model.lower()
        self.validate_threshold_bounds()
        self.validate_threshold_model()

    def get_mode(self):
        return "percentage" if "%" in self.lower_bound or "%" in self.upper_bound else "absolute"

    def validate_threshold_bounds(self):
        lower_bound = int(self.lower_bound.replace("%", ""))
        upper_bound = int(self.upper_bound.replace("%", ""))
        if lower_bound < 0 or upper_bound < 0:
            raise TableThresholdBoundsException("Threshold bounds for table cannot be negative.")
        if lower_bound > upper_bound:
            raise TableThresholdBoundsException("Lower bound cannot be greater than upper bound.")

    def validate_threshold_model(self):
        if self.model not in ["mismatch"]:
            raise InvalidModelForTableThreshold(
                f"Invalid model for Table Threshold: expected 'mismatch', but got '{self.model}'."
            )


@dataclass
class Filters:
    source: str | None = None
    target: str | None = None


def to_lower_case(input_list: list[str]) -> list[str]:
    return [element.lower() for element in input_list]


@dataclass
class Table:
    source_name: str
    target_name: str
    sampling_options: SamplingOptions | None = None
    aggregates: list[Aggregate] | None = None
    join_columns: list[str] | None = None
    jdbc_reader_options: JdbcReaderOptions | None = None
    select_columns: list[str] | None = None
    drop_columns: list[str] | None = None
    column_mapping: list[ColumnMapping] | None = None
    transformations: list[Transformation] | None = None
    column_thresholds: list[ColumnThresholds] | None = None
    filters: Filters | None = None
    table_thresholds: list[TableThresholds] | None = None

    def __post_init__(self):
        self.source_name = self.source_name.lower()
        self.target_name = self.target_name.lower()
        self.select_columns = to_lower_case(self.select_columns) if self.select_columns else None
        self.drop_columns = to_lower_case(self.drop_columns) if self.drop_columns else None
        self.join_columns = to_lower_case(self.join_columns) if self.join_columns else None

    @property
    def to_src_col_map(self):
        if self.column_mapping:
            return {c.source_name: c.target_name for c in self.column_mapping}
        return None

    @property
    def to_tgt_col_map(self):
        if self.column_mapping:
            return {c.target_name: c.source_name for c in self.column_mapping}
        return None

    def get_src_to_tgt_col_mapping_list(self, cols: list[str], layer: str) -> set[str]:
        if layer == "source":
            return set(cols)
        if self.to_src_col_map:
            return {self.to_src_col_map.get(col, col) for col in cols}
        return set(cols)

    def get_layer_src_to_tgt_col_mapping(self, column_name: str, layer: str) -> str:
        if layer == "source":
            return column_name
        if self.to_src_col_map:
            return self.to_src_col_map.get(column_name, column_name)
        return column_name

    def get_tgt_to_src_col_mapping_list(self, cols: list[str] | set[str]) -> set[str]:
        if self.to_tgt_col_map:
            return {self.to_tgt_col_map.get(col, col) for col in cols}
        return set(cols)

    def get_layer_tgt_to_src_col_mapping(self, column_name: str, layer: str) -> str:
        if layer == "source":
            return column_name
        if self.to_tgt_col_map:
            return self.to_tgt_col_map.get(column_name, column_name)
        return column_name

    def get_select_columns(self, schema: list[Schema], layer: str) -> set[str]:
        if self.select_columns is None:
            return {sch.column_name for sch in schema}
        if self.to_src_col_map:
            return self.get_src_to_tgt_col_mapping_list(self.select_columns, layer)
        return set(self.select_columns)

    def get_threshold_columns(self, layer: str) -> set[str]:
        if self.column_thresholds is None:
            return set()
        return {self.get_layer_src_to_tgt_col_mapping(thresh.column_name, layer) for thresh in self.column_thresholds}

    def get_join_columns(self, layer: str) -> set[str] | None:
        if self.join_columns is None:
            return None
        return {self.get_layer_src_to_tgt_col_mapping(col, layer) for col in self.join_columns}

    def get_drop_columns(self, layer: str) -> set[str]:
        if self.drop_columns is None:
            return set()
        return {self.get_layer_src_to_tgt_col_mapping(col, layer) for col in self.drop_columns}

    def get_partition_column(self, layer: str) -> set[str]:
        if self.jdbc_reader_options and layer == "source":
            if self.jdbc_reader_options.partition_column:
                return {self.jdbc_reader_options.partition_column}
        return set()

    def get_filter(self, layer: str) -> str | None:
        if self.filters is None:
            return None
        if layer == "source":
            return self.filters.source
        return self.filters.target


@dataclass
class Schema:
    # TODO remove: This will have the value of ansi_normalized_column_name. Kept for backwards compatibility.
    column_name: str
    data_type: str
    ansi_normalized_column_name: str
    source_normalized_column_name: str


@dataclass
class Aggregate:
    agg_columns: list[str]
    type: str
    group_by_columns: list[str] | None = None

    def __post_init__(self):
        self.agg_columns = to_lower_case(self.agg_columns)
        self.type = self.type.lower()
        self.group_by_columns = to_lower_case(self.group_by_columns) if self.group_by_columns else None
        assert (
            self.type in _SUPPORTED_AGG_TYPES
        ), f"Invalid aggregate type: {self.type}, only {_SUPPORTED_AGG_TYPES} are supported."

    def get_agg_type(self):
        return self.type

    @classmethod
    def _join_columns(cls, columns: list[str]):
        return "+__+".join([DialectUtils.unnormalize_identifier(col) for col in columns])

    @property
    def group_by_columns_as_str(self):
        return self._join_columns(self.group_by_columns) if self.group_by_columns else "NA"

    @property
    def agg_columns_as_str(self):
        return self._join_columns(self.agg_columns)


@dataclass
class AggregateRule:
    agg_type: str
    agg_column: str
    group_by_columns: list[str] | None
    group_by_columns_as_str: str
    rule_type: str = "AGGREGATE"

    @property
    def column_from_rule(self):
        # creates rule_column. e.g., min_col1_grp1_grp2
        return f"{self.agg_type}_{self.agg_column}_{self.group_by_columns_as_str}"

    @property
    def group_by_columns_as_table_column(self):
        # If group_by_columns are not defined, store is as null
        group_by_cols_as_table_col = "NULL"
        if self.group_by_columns:
            # Sort the columns, convert to lower case and create a string:  , e.g., grp1, grp2
            formatted_cols = ", ".join([f"{col.lower()}" for col in sorted(self.group_by_columns)])
            group_by_cols_as_table_col = f"\"{formatted_cols}\""
        return group_by_cols_as_table_col

    def get_rule_query(self, rule_id):
        rule_info = f""" map( 'agg_type', '{self.agg_type}',
                 'agg_column', '{self.agg_column}',
                 'group_by_columns', {self.group_by_columns_as_table_column}
                 )
        """
        return f" SELECT {rule_id} as rule_id, " f" '{self.rule_type}' as rule_type, " f" {rule_info} as rule_info "


@dataclass
class AggregateQueryRules:
    layer: str
    group_by_columns: list[str] | None
    group_by_columns_as_str: str
    query: str
    rules: list[AggregateRule]



from typing import cast

from databricks.labs.blueprint.installation import JsonList, JsonObject, RootJsonValue

from databricks.labs.lakebridge.config import LSPConfigOptionV1, LSPPromptMethod
from databricks.labs.lakebridge.transpiler.repository import TranspilerInfo, TranspilerRepository


class TranspilersDescription:
    def __init__(self, transpiler_repository: TranspilerRepository) -> None:
        self._transpiler_repository = transpiler_repository

    @classmethod
    def transpiler_as_json(cls, transpiler: TranspilerInfo) -> JsonObject:
        return {
            "name": transpiler.transpiler_name,
            "config-path": str(transpiler.configuration_path),
            "versions": {
                "installed": transpiler.version,
                # TODO: Determine the lastest version available, if possible.
                "latest": None,
            },
            "supported-dialects": {
                dialect: {"options": [cls.dialect_options_as_json(option) for option in options]}
                for dialect, options in transpiler.dialects.items()
            },
        }

    @classmethod
    def dialect_options_as_json(cls, option: LSPConfigOptionV1) -> JsonObject:
        description: JsonObject = {
            "flag": option.flag,
            "method": option.method.name,
        }
        if option.method != LSPPromptMethod.FORCE:
            description["prompt"] = option.prompt
        if option.method == LSPPromptMethod.CHOICE:
            description["choices"] = cast(JsonList, option.choices)
        if option.default is not None:
            description["default"] = option.default
        return description

    def as_json(self) -> RootJsonValue:
        """Obtain a JSON description of the installed transpilers.

        Information includes:
         - Plugins and their versions.
         - Dialects supported by each plugin.
         - The options that each dialect supports.
        """
        transpiler_repository = self._transpiler_repository
        installed_transpilers = transpiler_repository.installed_transpilers()
        all_dialects = cast(JsonList, sorted(list(transpiler_repository.all_dialects())))
        return {
            "installed-transpilers": [
                self.transpiler_as_json(info) for transpiler_id, info in sorted(installed_transpilers.items())
            ],
            "available-dialects": all_dialects,
        }



import io
import logging
import os
import random
import string
from datetime import datetime, timezone
from pathlib import Path

from databricks.labs.blueprint.installation import RootJsonValue
from databricks.sdk import WorkspaceClient

logger = logging.getLogger(__name__)


class SwitchRunner:
    """Runner for Switch LLM transpilation jobs."""

    def __init__(
        self,
        ws: WorkspaceClient,
    ):
        self._ws = ws

    def run(
        self,
        volume_input_path: str,
        output_ws_folder: str,
        source_tech: str,
        catalog: str,
        schema: str,
        foundation_model: str,
        job_id: int,
    ) -> RootJsonValue:
        """Trigger Switch job."""

        job_params = self._build_job_parameters(
            input_dir=volume_input_path,
            output_dir=output_ws_folder,
            source_tech=source_tech,
            catalog=catalog,
            schema=schema,
            foundation_model=foundation_model,
        )
        logger.info(f"Triggering Switch job with job_id: {job_id}")

        return self._run_job(job_id, job_params)

    def upload_to_volume(
        self,
        local_path: Path,
        catalog: str,
        schema: str,
        volume: str,
    ) -> str:
        """Upload local files to UC Volume with unique timestamped path."""
        now = datetime.now(timezone.utc)
        time_part = now.strftime("%Y%m%d%H%M%S")
        random_part = ''.join(random.choices(string.ascii_lowercase + string.digits, k=4))
        volume_base_path = f"/Volumes/{catalog}/{schema}/{volume}"
        volume_input_path = f"{volume_base_path}/input-{time_part}-{random_part}"

        logger.info(f"Uploading {local_path} to {volume_input_path}...")

        # File upload
        if local_path.is_file():
            if local_path.name.startswith('.'):
                logger.debug(f"Skipping hidden file: {local_path}")
                return volume_input_path
            volume_file_path = f"{volume_input_path}/{local_path.name}"
            with open(local_path, 'rb') as f:
                content = f.read()
            self._ws.files.upload(file_path=volume_file_path, contents=io.BytesIO(content), overwrite=True)
            logger.debug(f"Uploaded: {local_path} -> {volume_file_path}")

        # Directory upload
        else:
            for root, dirs, files in os.walk(local_path):
                # remove hidden directories
                dirs[:] = [d for d in dirs if not d.startswith('.')]
                # skip hidden files
                files = [f for f in files if not f.startswith('.')]
                for file in files:
                    local_file = Path(root) / file
                    relative_path = local_file.relative_to(local_path)
                    volume_file_path = f"{volume_input_path}/{relative_path}"

                    with open(local_file, 'rb') as f:
                        content = f.read()

                    self._ws.files.upload(file_path=volume_file_path, contents=io.BytesIO(content), overwrite=True)
                    logger.debug(f"Uploaded: {local_file} -> {volume_file_path}")

        logger.info(f"Upload complete: {volume_input_path}")
        return volume_input_path

    def _build_job_parameters(
        self,
        input_dir: str,
        output_dir: str,
        source_tech: str,
        catalog: str,
        schema: str,
        foundation_model: str,
        switch_options: dict | None = None,
    ) -> dict[str, str]:
        """Build Switch job parameters."""
        if switch_options is None:
            switch_options = {}
        return {
            "input_dir": input_dir,
            "output_dir": output_dir,
            "source_tech": source_tech,
            "catalog": catalog,
            "schema": schema,
            "foundation_model": foundation_model,
            **switch_options,
        }

    def _run_job(
        self,
        job_id: int,
        job_params: dict[str, str],
    ) -> RootJsonValue:
        """Trigger Switch job and return run information."""
        job_run = self._ws.jobs.run_now(job_id, job_parameters=job_params)

        if not job_run.run_id:
            raise SystemExit(f"Job {job_id} execution failed.")

        job_run_url = f"{self._ws.config.host}/jobs/{job_id}/runs/{job_run.run_id}"
        logger.info(f"Switch LLM transpilation job started: {job_run_url}")

        return [
            {
                "job_id": job_id,
                "run_id": job_run.run_id,
                "run_url": job_run_url,
            }
        ]



from __future__ import annotations

from dataclasses import dataclass
from enum import Enum
from pathlib import Path


# not using StrEnum because they only appear with Python 3.11
class ErrorSeverity(Enum):
    INFO = "INFO"
    WARNING = "WARNING"
    ERROR = "ERROR"


class ErrorKind(Enum):
    ANALYSIS = "ANALYSIS"
    PARSING = "PARSING"
    GENERATION = "GENERATION"
    VALIDATION = "VALIDATION"
    INTERNAL = "INTERNAL"


@dataclass
class CodePosition:
    line: int  # 0-based line number
    character: int  # 0-based character number


@dataclass
class CodeRange:
    start: CodePosition
    end: CodePosition


@dataclass
class TranspileError:
    code: str
    kind: ErrorKind
    severity: ErrorSeverity
    path: Path
    message: str
    range: CodeRange | None = None

    def __str__(self):
        return f"{type(self).__name__}(code={self.code}, kind={self.kind.name}, severity={self.severity.name}, path='{self.path!s}', message='{self.message}')"


@dataclass
class TranspileStatus:
    file_list: list[Path]
    no_of_transpiled_queries: int
    error_list: list[TranspileError]

    @property
    def analysis_error_count(self) -> int:
        return len([error for error in self.error_list if error.kind == ErrorKind.ANALYSIS])

    @property
    def parsing_error_count(self) -> int:
        return len([error for error in self.error_list if error.kind == ErrorKind.PARSING])

    @property
    def generation_error_count(self) -> int:
        return len([error for error in self.error_list if error.kind == ErrorKind.GENERATION])

    @property
    def validation_error_count(self) -> int:
        return len([error for error in self.error_list if error.kind == ErrorKind.VALIDATION])



import logging
import re
from datetime import datetime

from pyspark.errors import PySparkException
from pyspark.sql import DataFrame, SparkSession
from pyspark.sql.functions import col
from sqlglot import Dialect

from databricks.labs.lakebridge.reconcile.connectors.data_source import DataSource
from databricks.labs.lakebridge.reconcile.connectors.models import NormalizedIdentifier
from databricks.labs.lakebridge.reconcile.connectors.secrets import SecretsMixin
from databricks.labs.lakebridge.reconcile.connectors.dialect_utils import DialectUtils
from databricks.labs.lakebridge.reconcile.recon_config import JdbcReaderOptions, Schema
from databricks.sdk import WorkspaceClient

logger = logging.getLogger(__name__)


def _get_schema_query(catalog: str, schema: str, table: str):
    # TODO: Ensure that the target_catalog in the configuration is not set to "hive_metastore". The source_catalog
    #  can only be set to "hive_metastore" if the source type is "databricks".
    if schema == "global_temp":
        return f"describe table global_temp.{table}"
    if catalog == "hive_metastore":
        return f"describe table {catalog}.{schema}.{table}"

    query = f"""select
                            lower(column_name) as col_name,
                             full_data_type as data_type
                       from {catalog}.information_schema.columns
                       where lower(table_catalog)='{catalog}'
                                    and lower(table_schema)='{schema}'
                                     and lower(table_name) ='{table}'
                       order by col_name"""
    return re.sub(r'\s+', ' ', query)


class DatabricksDataSource(DataSource, SecretsMixin):
    _IDENTIFIER_DELIMITER = "`"

    def __init__(
        self,
        engine: Dialect,
        spark: SparkSession,
        ws: WorkspaceClient,
        secret_scope: str,
    ):
        self._engine = engine
        self._spark = spark
        self._ws = ws
        self._secret_scope = secret_scope

    def read_data(
        self,
        catalog: str | None,
        schema: str,
        table: str,
        query: str,
        options: JdbcReaderOptions | None,
    ) -> DataFrame:
        namespace_catalog = "hive_metastore" if not catalog else catalog
        if schema == "global_temp":
            namespace_catalog = "global_temp"
        else:
            namespace_catalog = f"{namespace_catalog}.{schema}"
        table_with_namespace = f"{namespace_catalog}.{table}"
        table_query = query.replace(":tbl", table_with_namespace)
        try:
            df = self._spark.sql(table_query)
            return df.select([col(column).alias(column.lower()) for column in df.columns])
        except (RuntimeError, PySparkException) as e:
            return self.log_and_throw_exception(e, "data", table_query)

    def get_schema(
        self,
        catalog: str | None,
        schema: str,
        table: str,
        normalize: bool = True,
    ) -> list[Schema]:
        catalog_str = catalog if catalog else "hive_metastore"
        schema_query = _get_schema_query(catalog_str, schema, table)
        try:
            logger.debug(f"Fetching schema using query: \n`{schema_query}`")
            logger.info(f"Fetching Schema: Started at: {datetime.now()}")
            schema_metadata = (
                self._spark.sql(schema_query)
                .selectExpr("col_name as column_name", "data_type")
                .where("column_name not like '#%'")
                .distinct()
                .collect()
            )
            logger.info(f"Schema fetched successfully. Completed at: {datetime.now()}")
            return [self._map_meta_column(field, normalize) for field in schema_metadata]
        except (RuntimeError, PySparkException) as e:
            return self.log_and_throw_exception(e, "schema", schema_query)

    def normalize_identifier(self, identifier: str) -> NormalizedIdentifier:
        return DialectUtils.normalize_identifier(
            identifier,
            source_start_delimiter=DatabricksDataSource._IDENTIFIER_DELIMITER,
            source_end_delimiter=DatabricksDataSource._IDENTIFIER_DELIMITER,
        )



import os
from collections.abc import Mapping

from databricks.sdk.useragent import alphanum_pattern, semver_pattern

# Entrypoint constants
DEFAULT_ENTRYPOINT = "cli"
VALID_ENTRYPOINTS = {DEFAULT_ENTRYPOINT, "desktop-app", "databricks-app"}


def make_alphanum_or_semver(value: str) -> str:
    if alphanum_pattern.match(value) or semver_pattern.match(value):
        return value
    # assume it's not a semver, replace illegal alphanum chars
    result = []
    for char in value:
        if not alphanum_pattern.match(char):
            char = '_'
        result.append(char)
    return "".join(result)


def get_entrypoint_from_env(environ: Mapping[str, str] | None = None) -> str:
    """Detect execution entrypoint from LAKEBRIDGE_ENTRYPOINT env var.

    Args:
        environ: Optional environment dict. If None, uses os.environ.

    Returns:
        One of "cli", "desktop-app", or "databricks-app". Defaults to "cli" if
        the environment variable is missing or contains an invalid value.
    """
    if environ is None:
        environ = os.environ
    entrypoint = environ.get("LAKEBRIDGE_ENTRYPOINT")
    if not entrypoint:
        return DEFAULT_ENTRYPOINT

    normalized = entrypoint.lower().strip()

    return normalized if normalized in VALID_ENTRYPOINTS else DEFAULT_ENTRYPOINT
